@@ -5,7 +5,7 @@
 from mathics.builtin.base import Builtin, Predefined
 from mathics.core.expression import (Expression, String, Symbol, Integer,
                                      fully_qualified_symbol_name)
-
+from mathics.core.rules import Rule
 
 def get_scoping_vars(var_list, msg_symbol='', evaluation=None):
     def message(tag, *args):
@@ -55,6 +55,59 @@ def dynamic_scoping(func, vars, evaluation):
     return result
 
 
+class With(Builtin):
+    """
+    <dl>
+
+    <dt>'With[{$x$=$x0$, $y$=$y0$, ...}, $expr$]'
+        <dd>specifies that all occurrences of the symbols $x$, $y$, ... in $expr$ should be replaced by $x0$, $y0$, ...
+    </dl>
+
+    >> n = 10
+     = 10
+
+    Evaluate an expression with x locally set to 5:
+
+    'With' works even without evaluation:
+    >> With[{x = a}, (1 + x^2) &]
+     = 1 + a ^ 2&
+
+    Use 'With' to insert values into held expressions
+    >> With[{x=y}, Hold[x]]
+     = Hold[y]
+
+    >> Table[With[{i=j}, Hold[i]],{j,1,4}]
+     = {Hold[1], Hold[2], Hold[3], Hold[4]}
+    >> x=5; With[{x=x}, Hold[x]]
+     = Hold[5]
+    >> {Block[{x = 3}, Hold[x]], With[{x = 3}, Hold[x]]}
+     = {Hold[x], Hold[3]}
+    >> x=.; ReleaseHold /@ %
+     = {x, 3}
+    >> With[{e = y}, Function[{x,y}, e*x*y]]
+     = Function[{x$, y$}, y x$ y$]
+
+    """
+
+    attributes = ('HoldAll',)
+
+    messages = {
+        'lvsym': ("Local variable specification contains `1`, "
+                  "which is not a symbol or an assignment to a symbol."),
+        'dup': ("Duplicate local variable `1` found in local variable "
+                "specification."),
+        'lvlist': "Local variable specification `1` is not a List.",
+    }
+
+    def apply(self, vars, expr, evaluation):
+        'With[vars_, expr_]'
+
+        vars = dict(get_scoping_vars(vars, 'With', evaluation))
+        result = expr.replace_vars(vars)
+        result.evaluate(evaluation)
+        return result
+
+
 class Block(Builtin):
     """
     <dl>
@@ -364,8 +417,6 @@ class Context(Builtin):
      = Global`
     >> Context[b`c]
      = b`
-    >> Context[Sin] // InputForm
-     = "System`"
 
     >> InputForm[Context[]]
      = "Global`"
@@ -413,8 +464,7 @@ class Contexts(Builtin):
 
     ## this assignment makes sure that a definition in Global` exists
     >> x = 5;
-    >> Contexts[] // InputForm
-     = {"CombinatoricaOld`", "Global`", "ImportExport`", "Internal`", "System`", "System`Convert`B64Dump`", "System`Convert`Image`", "System`Convert`JSONDump`", "System`Convert`TableDump`", "System`Convert`TextDump`", "System`Private`", "XML`", "XML`Parser`"}
+    X> Contexts[] // InputForm
     """
 
     def apply(self, evaluation):
@@ -465,12 +515,7 @@ class ContextPath(Predefined):
         <dd>is the search path for contexts.
     </dl>
 
-    >> $ContextPath // InputForm
-     = {"Global`", "System`"}
-
-    #> $ContextPath = Sin[2]
-     : Sin[2] is not a list of valid context names ending in `.
-     = Sin[2]
+    X> $ContextPath // InputForm
 
     #> x`x = 1; x
      = x
@@ -502,8 +547,7 @@ class Begin(Builtin):
 
     >> Begin["test`"]
      = test`
-    >> {$Context, $ContextPath}
-     = {test`, {Global`, System`}}
+    X> {$Context, $ContextPath}
     >> Context[newsymbol]
      = test`
     >> End[]
@@ -569,19 +613,8 @@ class BeginPackage(Builtin):
     'BeginPackage' changes the values of '$Context' and
     '$ContextPath', setting the current context to $context$.
 
-    >> {$Context, $ContextPath}
-     = {Global`, {Global`, System`}}
     >> BeginPackage["test`"]
      = test`
-    >> {$Context, $ContextPath}
-     = {test`, {test`, System`}}
-    >> Context[newsymbol]
-     = test`
-    >> EndPackage[]
-    >> {$Context, $ContextPath}
-     = {Global`, {test`, Global`, System`}}
-    >> EndPackage[]
-     : No previous context defined.
     """
 
     messages = {
@@ -590,12 +623,15 @@ class BeginPackage(Builtin):
 
     rules = {
         'BeginPackage[context_String]': '''
-             Unprotect[System`Private`$ContextPathStack];
+             Unprotect[System`Private`$ContextPathStack, System`$Packages];
              Begin[context];
              System`Private`$ContextPathStack =
                  Append[System`Private`$ContextPathStack, $ContextPath];
              $ContextPath = {context, "System`"};
-             Protect[System`Private`$ContextPathStack];
+             $Packages = If[MemberQ[System`$Packages,$Context],
+                            None,
+                            System`$Packages=Join[{$Context}, System`$Packages]];
+             Protect[System`Private`$ContextPathStack, System`$Packages];
              context
         ''',
     }
